
"""
Contact Page
"""
from shaft import (Shaft, page_meta, get_config, abort, get, post,
                    flash_success, flash_error, flash_data, get_flash_data,
                   register_package, redirect, request, url_for, send_mail,
                   recaptcha, nav_menu, route)
from shaft.exceptions import AppError
import shaft.utils as utils
import logging

# app_data
app_data = None
try:
    from shaft.contrib.app_data import AppData
    app_data = AppData("ContactPage")
except ImportError as e:
    pass


register_package(__package__)

__version__ = "1.0.0"

def main(**kwargs):
    """
        - config:
            - send_to
            - return_to
            - title
            - success_message
    """

    navigation = kwargs.get("nav_menu", {})
    navigation.setdefault("title", "Contact")
    navigation.setdefault("visible", True)
    navigation.setdefault("order", 100)

    options = kwargs.get("options", {})

    if app_data:
        if options.get("recipients") \
                and not app_data.has("recipients"):
            app_data.set(key="recipients",
                         value=options.get("recipients"),
                         description="Contact page recipients. Separated by comma")


    class ContactPage(Shaft):
        base_route = "/"
        decorators = Shaft.decorators + kwargs.get("decorators")

        @nav_menu(**navigation)
        @get("contact/")
        @post("contact/")
        def index(self):

            # Email to
            if app_data and app_data.get("recipients"):
                send_to = app_data.get("recipients")
            else:
                send_to = options.get("recipients", get_config("CONTACT_EMAIL", None))

            success_message = options.get("success_message", "Message sent. Thank you!")

            return_to = options.get("return_to", None)
            if return_to:
                if "/" not in return_to:
                    return_to = url_for(return_to)
            else:
                return_to = url_for(self)

            if not send_to:
                abort(500, "ContactPage missing email recipient")

            if request.method == "POST":
                email = request.form.get("email")
                subject = request.form.get("subject")
                message = request.form.get("message")
                name = request.form.get("name")

                try:
                    if recaptcha.verify():
                        if not email or not subject or not message:
                            raise AppError("All fields are required")
                        elif not utils.is_email_valid(email):
                            raise AppError("Invalid email address")
                        else:
                            try:
                                send_mail(to=send_to,
                                          reply_to=email,
                                          mail_from=email,
                                          mail_subject=subject,
                                          mail_message=message,
                                          mail_name=name,
                                          template=options.get("template", "contact-us.txt")
                                          )
                                flash_data("ContactPage:EmailSent")
                            except Exception as ex:
                                logging.exception(ex)
                                raise AppError("Unable to send email")
                    else:
                        raise AppError("Security code is invalid")
                except AppError as e:
                    flash_error(e.message)
                return redirect(self)

            title = options.get("title", "Contact Us")
            page_meta(title=title)

            fd = get_flash_data()
            return {
                "title": title,
                "email_sent": True if fd and "ContactPage:EmailSent" in fd else False,
                "success_message": success_message,
                "return_to": return_to
            }

    return ContactPage





"""
Some common functions
"""
from __future__ import division
import os
import re
import inspect
import time
import datetime
import string
import random
import socket
import itsdangerous
import humanize
import pkg_resources
import urllib
import hashlib
import json
import uuid
from . import exceptions
from slugify import slugify
from werkzeug.utils import import_string
from passlib.hash import bcrypt as crypt_engine
from distutils.dir_util import (copy_tree as copy_dir,
                                remove_tree as remove_dir,
                                mkpath as make_dirs)
from distutils.file_util import copy_file, move_file
from inflection import (dasherize,
                        underscore,
                        camelize,
                        pluralize,
                        singularize,
                        titleize)
from six.moves.urllib.parse import (urlparse,
                                    urlencode,
                                    unquote_plus as urllib_unquote_plus)




def is_email_valid(email):
    """
    Check if email is valid
    """
    pattern = re.compile(r'[\w\.-]+@[\w\.-]+[.]\w+')
    return bool(pattern.match(email))


def is_password_valid(password):
    """
    Check if a password is valid
    """
    pattern = re.compile(r"^.{4,75}$")
    return bool(pattern.match(password))


def is_url_valid(url):
    """
    Check if url is valid
    """
    pattern = re.compile(
            r'^(?:http|ftp)s?://' # http:// or https://
            r'(?:(?:[A-Z0-9](?:[A-Z0-9-]{0,61}[A-Z0-9])?\.)+(?:[A-Z]{2,6}\.?|[A-Z0-9-]{2,}\.?)|' #domain...
            #r'localhost|' #localhost...
            r'\d{1,3}\.\d{1,3}\.\d{1,3}\.\d{1,3})' # ...or ip
            r'(?::\d+)?' # optional port
            r'(?:/?|[/?]\S+)$', re.IGNORECASE)
    return bool(pattern.match(url))


def urldecode(s):
    return urllib_unquote_plus(s)


def md5(value):
    """
    Create MD5
    :param value:
    :return:
    """
    m = hashlib.md5()
    m.update(value)
    return str(m.hexdigest())


def guid():
    """
    Creates and returns a UUID 4 hex value
    :return: string
    """
    return uuid.uuid4().hex


def chunk_list(items, size):
    """
    Return a list of chunks
    :param items: List
    :param size: int The number of items per chunk
    :return: List
    """
    size = max(1, size)
    return [items[i:i + size] for i in range(0, len(items), size)]


def in_any_list(items1, items2):
    """
    Check if any items are in list 3
    :param items1: list
    :param items2: list
    :return:
    """
    return any(i in items2 for i in items1)


def encrypt_string(string):
    """
    Encrypt a string
    """
    return crypt_engine.encrypt(string)


def verify_encrypted_string(string, encrypted_string):
    """
    Verify an encrypted string
    """
    return crypt_engine.verify(string, encrypted_string)


def generate_random_string(length=8):
    """
    Generate a random string
    """
    char_set = string.ascii_uppercase + string.digits
    return ''.join(random.sample(char_set * (length - 1), length))


def generate_random_hash(size=32):
    """
    Return a random hash key
    :param size: The max size of the hash
    :return: string
    """
    return os.urandom(size//2).encode('hex')


def how_old(dob):
    """
    Calculate the age
    :param dob: datetime object
    :return: int
    """
    today = datetime.date.today()
    return today.year - dob.year - ((today.month, today.day) < (dob.month, dob.day))


def is_port_open(port, host="127.0.0.1"):
    """
    Check if a port is open
    :param port:
    :param host:
    :return bool:
    """
    s = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
    try:
        s.connect((host, int(port)))
        s.shutdown(2)
        return True
    except Exception as e:
        return False

def create_uri(host, user=None, password=None, port=None, path=None, scheme="http"):
    port = ":%s" % port if port else ""
    path = "/%s" % path if path else ""
    user_info = ""
    if user or password:
        user_info = "%s:%s@" % (user or "", password or "")
    return "{scheme}://{user_info}{host}{port}{path}".format(
        scheme=scheme, host=host, user_info=user_info,port=port, path=path)


class dictdot(dict):
    """
    A dict extension that allows dot notation to access the data.
    ie: dict.get('key.key2.0.keyx')
    my_dict = {...}
    d = dictdot(my_dict)
    d.get("key1")
    d.get("key1.key2")
    d.get("key3.key4.0.keyX")

    Still have the ability to access it as a normal dict
    d[key1][key2]
    """
    def get(self, key, default=None):
        """
        Access data via
        :param key:
        :param default: the default value
        :return:
        """
        try:
            val = self
            if "." not in key:
                return self[key]
            for k in key.split('.'):
                if k.isdigit():
                    k = int(k)
                val = val[k]
            return val
        except (TypeError, KeyError, IndexError) as e:
            return default


def sign_jwt(data, secret_key, expires_in, salt=None, **kw):
    """
    To create a signed JWT
    :param data:
    :param secret_key:
    :param expires_in:
    :param salt:
    :param kw:
    :return: string
    """
    try:
        s = itsdangerous.TimedJSONWebSignatureSerializer(secret_key=secret_key,
                                                         expires_in=expires_in,
                                                         salt=salt,
                                                          **kw)
        return s.dumps(data)

    except itsdangerous.BadData as e:
        raise exceptions.ExtensionError(e)

def unsign_jwt(token, secret_key, salt=None, **kw):
    """
    To unsign a JWT token
    :param token:
    :param kw:
    :return: mixed data
    """
    try:
        s = itsdangerous.TimedJSONWebSignatureSerializer(secret_key, salt=salt, **kw)
        return s.loads(token)
    except itsdangerous.BadData as e:
        raise exceptions.ExtensionError(e)


class TimestampSigner2(itsdangerous.TimestampSigner):
    expires_in = 0

    def get_timestamp(self):
        now = datetime.datetime.utcnow()
        expires_in = now + datetime.timedelta(seconds=self.expires_in)
        return int(expires_in.strftime("%s"))

    @staticmethod
    def timestamp_to_datetime(ts):
        return datetime.datetime.fromtimestamp(ts)


class URLSafeTimedSerializer2(itsdangerous.URLSafeTimedSerializer):
    default_signer = TimestampSigner2

    def __init__(self, secret_key, expires_in=3600, salt=None, **kwargs):
        self.default_signer.expires_in = expires_in
        super(self.__class__, self).__init__(secret_key, salt=salt, **kwargs)


def sign_url_safe(data, secret_key, expires_in=None, salt=None, **kw):
    """
    To sign url safe data.
    If expires_in is provided it will Time the signature
    :param data:
    :param secret_key:
    :param expires_in:
    :param salt:
    :param kw:
    :return:
    """
    try:
        if expires_in:
            s = URLSafeTimedSerializer2(secret_key=secret_key,
                                        expires_in=expires_in,
                                        salt=salt,
                                        **kw)
        else:
            s = itsdangerous.URLSafeSerializer(secret_key=secret_key,
                                               salt=salt,
                                               **kw)
        return s.dumps(data)
    except itsdangerous.BadData as e:
        raise exceptions.ExtensionError(e)


def unsign_url_safe(token, secret_key, salt=None, **kw):
    """
    To sign url safe data.
    If expires_in is provided it will Time the signature
    :param token:
    :param secret_key:
    :param salt:
    :param kw:
    :return:
    """
    try:
        if len(token.split(".")) == 3:
            s = URLSafeTimedSerializer2(secret_key=secret_key, salt=salt, **kw)
            value, timestamp = s.loads(token, max_age=None, return_timestamp=True)
            now = datetime.datetime.utcnow()
            if timestamp > now:
                return value
            else:
                raise itsdangerous.SignatureExpired(
                    'Signature age %s < %s ' % (timestamp, now),
                    payload=value,
                    date_signed=timestamp)
        else:
            s = itsdangerous.URLSafeSerializer(secret_key=secret_key, salt=salt, **kw)
            return s.loads(token)
    except itsdangerous.BadData as e:
        raise exceptions.ExtensionError(e)


# ------------------------------------------------------------------------------

def to_json(d):
    """
    Convert data to json. It formats datetime
    :param d: dict or list
    :return: json data
    """
    return json.dumps(d, cls=_DatetimeEncoder)


class _DatetimeEncoder(json.JSONEncoder):
    def default(self, obj):
        if isinstance(obj, datetime.datetime):
            return obj.strftime('%Y-%m-%dT%H:%M:%SZ')
        elif isinstance(obj, datetime.date):
            return obj.strftime('%Y-%m-%d')
        # Let the base class default method raise the TypeError
        return json.JSONEncoder.default(self, obj)




class InspectDecoratorCompatibilityError(Exception):
    pass


class _InspectMethodsDecorators(object):
    """
    This class attempt to retrieve all the decorators in a method
    """
    def __init__(self, method):
        self.method = method
        self.decos = []

    def parse(self):
        """
        Return the list of string of all the decorators found
        """
        self._parse(self.method)
        return list(set([deco for deco in self.decos if deco]))

    @classmethod
    def extract_deco(cls, line):
        line = line.strip()
        if line.startswith("@"):
            if "(" in line:
                line = line.split('(')[0].strip()
            return line.strip("@")

    def _parse(self, method):
        argspec = inspect.getargspec(method)
        args = argspec[0]
        if args and args[0] == 'self':
            return argspec
        if hasattr(method, '__func__'):
            method = method.__func__
        if not hasattr(method, '__closure__') or method.__closure__ is None:
            raise InspectDecoratorCompatibilityError

        closure = method.__closure__
        for cell in closure:
            inner_method = cell.cell_contents
            if inner_method is method:
                continue
            if not inspect.isfunction(inner_method) \
                and not inspect.ismethod(inner_method):
                continue
            src = inspect.getsourcelines(inner_method)[0]
            self.decos += [self.extract_deco(line) for line in src]
            self._parse(inner_method)


def get_decorators_list(method):
    """
    Shortcut to InspectMethodsDecorators
    :param method: object
    :return: List
    """
    kls = _InspectMethodsDecorators(method)
    return kls.parse()







"""
Shaft Auth
"""

import datetime
import flask_login
import signals
import exceptions
import logging
from flask_login import current_user
from flask import current_app
from shaft.exceptions import AppError
from shaft.core import get_installed_app_options
from shaft import (register_package,
                   get_config,
                   abort,
                   send_mail,
                   url_for,
                   views,
                   models,
                   utils,
                   request,
                   redirect,
                   flash,
                   session,
                   init_app)
from itsdangerous import (TimedJSONWebSignatureSerializer,
                          BadData)


register_package(__package__)

__version__ = "1.0.0"

ROLES_SUPERADMIN = ["SUPERADMIN"]
ROLES_ADMIN = ROLES_SUPERADMIN + ["ADMIN"]
ROLES_MANAGER = ROLES_ADMIN + ["MANAGER"]
ROLES_CONTRIBUTOR = ROLES_MANAGER + ["EDITOR", "CONTRIBUTOR"]
ROLES_MODERATOR = ROLES_CONTRIBUTOR + ["MODERATOR"]

JWT_TTL = 3600
JWT_SALT = "shaft.contrib.auth"


TOKENIZATION_SALTS = {
    "verify-email": "auth:verify-email",
    "reset-password": "auth:reset-password",
    "email": "auth:email"
}


_app_options = None


def main(app, **kwargs):

    @app.before_request
    def force_password_change():
        _ = _get_app_options().get("require_password_change_exclude_endpoints")
        _ = [] if not isinstance(_, list) else _

        exclude_endpoints = ["static", "ContactPage:index", "Index:index",
                             "AuthLogin:logout"] + _

        if current_user and current_user.is_authenticated:
            if request.endpoint \
                    and request.endpoint not in exclude_endpoints:
                if request.endpoint != "AuthAccount:change_password" \
                        and session_get_require_password_change():
                    flash("Password Change is required", "info")
                    return redirect(views.AuthAccount.change_password)


def _get_app_options():
    global _app_options
    if not _app_options:
        _app_options = get_installed_app_options(__name__)
    return _app_options


def is_authenticated():
    """ A shortcut to check if a user is authenticated """
    return current_user.is_authenticated


def not_authenticated():
    """ A shortcut to check if user not authenticated."""
    return not is_authenticated()


def get_random_password(length=8):
    return utils.generate_random_string(length)


# ------------------------------------------------------------------------------
#
def visible_to(*roles):
    """
    This is a @nav_menu specific function to set the visibility of menu based on
    roles
    :param roles:
    :return: callback fn
    """
    if is_authenticated():
        return True if current_user.has_any_roles(*roles) else False
    return False


# Alias to visible_to
visible_to_superadmins = lambda: visible_to(*ROLES_SUPERADMIN)
visible_to_admins = lambda: visible_to(*ROLES_ADMIN)
visible_to_managers = lambda: visible_to(*ROLES_MANAGER)
visible_to_contributors = lambda: visible_to(*ROLES_CONTRIBUTOR)
visible_to_moderators = lambda: visible_to(*ROLES_MODERATOR)
visible_to_authenticated = lambda: is_authenticated()
visible_to_notauthenticated = lambda: not_authenticated()

# ------------------------------------------------------------------------------

# TOKENIZATION


def get_jwt_secret():
    secret_key = get_config("JWT_SECRET") or get_config("SECRET_KEY")
    if not secret_key:
        raise exceptions.AuthError("Missing config JWT/SECRET_KEY")
    return secret_key


def make_jwt(user, expires_in=None):
    """
    Create a secure timed JWT token that can be passed. It save the user id,
    which later will be used to retrieve the data

    :param user: AuthUser, the user's object
    :param expires_in: - time in second for the token to expire
    :return: string
    """
    secret_key = get_jwt_secret()

    s = utils.sign_jwt(data={"id": user.id},
                       secret_key=secret_key,
                       salt=JWT_SALT,
                       expires_in=expires_in or JWT_TTL)
    return s


def get_jwt(token=None):
    """
    Return the user associated to the token, otherwise it will return None.
    If token is not provided, it will pull it from the headers: Authorization

    Exception:
    Along with AuthError, it may
    :param token:
    :return: User
    """

    if not token:
        if not 'Authorization' in request.headers:
            raise exceptions.AuthError("Missing Authorization Bearer in headers")
        data = request.headers['Authorization'].encode('ascii', 'ignore')
        token = str.replace(str(data), 'Bearer ', '').strip()

    secret_key = get_jwt_secret()

    s = utils.unsign_jwt(token=token,
                         secret_key=secret_key,
                         salt=JWT_SALT)
    if "id" not in s:
        raise exceptions.AuthError("Invalid Authorization Bearer Token")

    return get_user(int(s["id"]))


def make_user_url_safe(user, expires_in, salt):
    """
    Make an auth url safe token. usually to do some account action. It will
    connect to the UserLogin table
    It will set the user
    :param user: AuthUser
    :param expires_in:
    :param salt:
    :return:
    """
    secret_key = get_jwt_secret()
    return utils.sign_url_safe({"id": user.id},
                               secret_key=secret_key,
                               salt=TOKENIZATION_SALTS[salt],
                               expires_in=expires_in)


def get_user_url_safe(token, salt):
    """
    Get the user and return the user object .
    :param token:
    :param salt:
    :return: int : User
    """
    secret_key = get_jwt_secret()
    data = utils.unsign_url_safe(token,
                                 secret_key=secret_key,
                                 salt=TOKENIZATION_SALTS[salt])
    if "id" not in data:
        raise exceptions.AuthError("Invalid Token")
    return get_user(int(data["id"]))


def make_email_url_safe(email):
    """
    Create an email safe url
    :param email:
    :return:
    """
    secret_key = get_jwt_secret()
    return utils.sign_url_safe(email,
                               secret_key=secret_key,
                               salt=TOKENIZATION_SALTS["email"])


def get_email_url_safe(email):
    """
    Retrieve an url safe email
    :param email:
    :return:
    """
    secret_key = get_jwt_secret()
    return utils.unsign_url_safe(email,
                                 secret_key=secret_key,
                                 salt=TOKENIZATION_SALTS["email"])


# ------------------------------------------------------------------------------
# SIGNUP + LOGIN


def get_user(id):
    """
    To get a user by id
    :param id:
    :return: AuthUser object
    """
    return models.AuthUser.get(id)


def get_user_login(id):
    """
    To get a user by id
    :param id:
    :return: AuthUserLogin object
    """
    return models.AuthUserLogin.get(id)


def get_user_by_login_email(email):
    """
    Get the user by login email
    :param email:
    :return:  AuthUser
    """
    userl = models.AuthUserLogin.get_by_email(email)
    if userl:
        return userl.user
    return None


def login(email, password, require_verified_email=False, session=True):
    """
    To login with email and password
    Additionally, it may require_th
    :param email:
    :param password:
    :param require_verified_email:
    :param session:
    :return: on_login signal
    """
    def cb():
        userl = models.AuthUserLogin.get_by_email(email)
        if userl and userl.password_hash \
                and userl.password_matched(password):
            if require_verified_email and not userl.email_verified:
                raise exceptions.VerifyEmailError()
            user = userl.user
            user.update(last_login=datetime.datetime.now())
            if session:
                flask_login.login_user(user)
            return user
        return None
    return signals.on_login(cb)


def register(email, password, name):
    """
    To create a new email login
    :param email:
    :param password:
    :param name:
    :return: AuthUser
    """
    def cb():
        user_login = models\
            .AuthUserLogin.new(login_type=models.AuthUserLogin.TYPE_EMAIL,
                               email=email,
                               password=password.strip(),
                               user_info={
                                    "name": name,
                                    "contact_email": email
                               })
        return user_login.user
    return signals.on_register(cb)


def change_login(user, email):
    """
    Change user's login email
    :param user:
    :param email:
    :return:
    """
    def cb():
        user_login = user.get_login()
        return user_login.change_email(email)
    return signals.on_login_changed(cb)


def change_password(user, password, password_confirm=None):
    """
    Change a user's password
    :param user:
    :param password:
    :param password_confirm:
    :return:
    """
    def cb():
        user_login = user.get_login()
        if not utils.is_password_valid(password):
            raise exceptions.AuthError("Invalid Password")
        if password_confirm and password != password_confirm:
            raise exceptions.AuthError("Passwords don't match")
        return user_login.change_password(password)
    return signals.on_password_changed(cb)


def reset_password(user):
    """
    Return the new random password that has been reset
    :param user_login: AuthUserLogin
    :return: string - the new password
    """
    def cb():
        user_login = user.get_login()
        return user_login.change_password(get_random_password())
    return signals.on_password_reset(cb)

# ------------------------------------------------------------------------------
# EMAIL SENDING


def _url_for_email(endpoint, base_url=None, **kw):
    """
    Create an external url_for by using a custom base_url different from the domain we
    are on
    :param endpoint:
    :param base_url:
    :param kw:
    :return:
    """
    base_url = base_url or get_config("MAIL_EXTERNAL_BASE_URL")
    _external = True if not base_url else False
    url = url_for(endpoint, _external=_external, **kw)
    if base_url and not _external:
        url = "%s/%s" % (base_url.strip("/"), url.lstrip("/"))
    return url


def send_password_reset_email(user=None,
                              email=None,
                              base_url=None,
                              template=None,
                              method=None,
                              **kw):
    """
    Reset a password and send email
    :param user: AuthUser
    :param email: str - The auth user login email
    :param base_url: str - By default it will use the current url, base_url will allow custom url
    :param template: str - The email template
    :param method: str - token or email - The method to reset the password
    :param kwargs: Any data to pass
    :return:
    """

    if email and not user:
        user = get_user_by_login_email(email)
    if not user:
        raise exceptions.AuthError("Invalid account. Unable to send email")
    user_login = user.get_login()
    if user_login.login_type != models.AuthUserLogin.TYPE_EMAIL:
        raise exceptions.AuthError("Invalid login type. Must be the type of email to be sent email to")

    options = _get_app_options()
    method = method or options.get("reset_password_method") or "TOKEN"
    email_template = template \
                     or options.get("reset_password_email_template") \
                     or "reset-password.txt"
    new_password = None

    if method.upper() == "TOKEN":
        # in minutes
        expires_in = get_config("reset_password_token_ttl") or 60
        expires_in = expires_in * 60

        token = make_user_url_safe(user,
                                   expires_in=expires_in,
                                   salt="reset-password")

        email = make_email_url_safe(user_login.email)
        url = _url_for_email(views.AuthLogin.reset_password,
                             base_url=base_url,
                             token=token,
                             email=email)
    else:
        new_password = reset_password(user)
        url = _url_for_email(views.AuthLogin.login, base_url=base_url)

    send_mail(template=email_template,
              to=user_login.email,
              user=user.email_data,
              action={
                  "reset_method": method.upper(),
                  "url": url,
                  "new_password": new_password
              },
              data=kw)


def send_email_verification_email(user, base_url=None, template=None, **kw):

    options = _get_app_options()
    email_template = template \
                     or options.get("verify_email_template") \
                     or "verify-email.txt"

    user_login = user.get_login()
    url = _create_verify_email_token_url(user, base_url=base_url)

    send_mail(template=email_template,
              to=user_login.email,
              user=user.email_data,
              action={
                  "url": url
              },
              data=kw)


def send_registration_welcome_email(user, base_url=None, template=None, **kw):

    options = _get_app_options()
    verify_email = options.get("verify_email") or False
    email_template = template \
                     or options.get("verify_register_email_template") \
                     or "verify-register-email.txt"

    user_login = user.get_login()
    url = _create_verify_email_token_url(user, base_url=base_url)

    send_mail(template=email_template,
              to=user_login.email,
              user=user.email_data,
              action={
                  "require_email_verification": verify_email,
                  "url": url
              },
              data=kw)


def _create_verify_email_token_url(user, base_url=None):

    options = _get_app_options()
    expires_in = options.get("verify_email_token_ttl") or (60 * 24)

    user_login = user.get_login()
    user_login.set_email_verified(False)
    token = make_user_url_safe(user,
                               expires_in=expires_in,
                               salt="verify-email")
    email = make_email_url_safe(user_login.email)
    return _url_for_email(views.AuthLogin.verify_email,
                          token=token,
                          email=email,
                          base_url=base_url)


def session_set_require_password_change(change=True):
    session["auth:require_password_change"] = change


def session_get_require_password_change():
    return session.get("auth:require_password_change")


# ------------------------------------------------------------------------------
# CLI
from shaft.cli import CLI


class AuthCLI(CLI):
    def __init__(self, command, click):
        @command("auth:create-super-admin")
        @click.argument("email")
        def create_super_admin(email):
            """
            To create a super admin by providing the email address
            """
            print("-" * 80)
            print("Shaft Auth: Create Super Admin")
            print("Email: %s" % email)
            try:
                password = get_random_password()
                user = register(email=email, password=password, name="SuperAdmin")
                user.set_role(models.AuthRole.SUPERADMIN)
                user.update(require_password_change=True)

                print("Password: %s" % password)
            except Exception as e:
                print("ERROR: %s" % e)

            print("Done!")

        @command("auth:reset-password")
        @click.argument("email")
        def reset_password(email):
            """
            To reset password by email
            """
            print("-" * 80)
            print("Shaft Auth: Reset Password")
            try:
                ul = models.AuthUserLogin.get_by_email(email)

                if not ul:
                    raise Exception("Email '%s' doesn't exist" % email)
                print(ul.email)
                password = get_random_password()
                ul.change_password(password)
                ul.update(require_password_change=True)
                print("Email: %s" % email)
                print("New Password: %s" % password)
            except Exception as e:
                print("ERROR: %s" % e)

            print("Done!")

        @command("auth:user-info")
        @click.option("--email")
        @click.option("--id")
        def reset_password(email=None, id=None):
            """
            Get the user info by email or ID
            """
            print("-" * 80)
            print("Shaft Auth: User Info")
            print("")
            try:
                if email:
                    ul = models.AuthUserLogin.get_by_email(email)
                    if not ul:
                        raise Exception("Invalid Email address")
                    user_info = ul.user
                elif id:
                    user_info = models.AuthUser.get(id)
                    if not user_info:
                        raise Exception("Invalid User ID")

                k = [
                    ("ID", "id"), ("Name", "name"), ("First Name", "first_name"),
                    ("Last Name", "last_name"), ("Signup", "created_at"),
                    ("Last Login", "last_login"), ("Signup Method", "register_method"),
                    ("Is Active", "is_active")
                ]
                print("Email: %s" % user_info.get_email_login().email)
                for _ in k:
                    print("%s : %s" % (_[0], getattr(user_info, _[1])))

            except Exception as e:
                print("ERROR: %s" % e)

            print("")
            print("Done!")


# ---

from .decorators import *



#-------
# TODO: DEPRECATION


# DEPRECATED
def make_user_login_url_safe(user_login_id, expires_in, salt):
    """
    Make an auth url safe token. usually to do some account action. It will
    connect to the UserLogin table
    It will set the user
    :param user_login_id:
    :param expires_in:
    :param salt:
    :return:
    """
    secret_key = get_jwt_secret()
    data = {"id": user_login_id}
    return utils.sign_url_safe(data,
                               secret_key=secret_key,
                               salt=TOKENIZATION_SALTS[salt],
                               expires_in=expires_in)

# DEPRECATED
def get_user_login_url_safe(token, salt):
    """
    Get the user and return the user object .
    :param token:
    :param salt:
    :return: int : User
    """
    secret_key = get_jwt_secret()
    data = utils.unsign_url_safe(token,
                                 secret_key=secret_key,
                                 salt=TOKENIZATION_SALTS[salt])
    if "id" not in data:
        raise exceptions.AuthError("Invalid Token")
    return get_user_login(int(data["id"]))

# DEPRECATED
def login_with_email(email, password, require_verified_email=False, session=True):
    """
    To login with email and password
    Additionally, it may require_th
    :param email:
    :param password:
    :param require_verified_email:
    :param session:
    :return: on_login signal
    """
    def cb():
        userl = models.AuthUserLogin.get_by_email(email)
        if userl and userl.password_hash \
                and userl.password_matched(password):
            if require_verified_email and not userl.email_verified:
                raise exceptions.VerifyEmailError()
            userl.user.update(last_login=datetime.datetime.now())
            if session:
                flask_login.login_user(userl.user)
            return userl
        return None
    return signals.on_login(cb)


import functools
import inspect
import copy
from flask import Response, jsonify, request, current_app, url_for, make_response
from werkzeug.wrappers import BaseResponse
from jinja2 import Markup
from dicttoxml import dicttoxml
from .core import Shaft, init_app, apply_function_to_members
from . import ext
from . import utils
import blinker
import flask_cors


def route(rule=None, **kwargs):
    """
    This decorator defines custom route for both class and methods in the view.
    It behaves the same way as Flask's @app.route

    on class:
        It takes the following args
            - rule: the root route of the endpoint
            - decorators: a list of decorators to run on each method

    on methods:
        along with the rule, it takes kwargs
            - endpoint
            - defaults
            - ...

    :param rule:
    :param kwargs:
    :return:
    """

    _restricted_keys = ["extends", "route", "decorators"]
    def decorator(f):
        if inspect.isclass(f):
            extends = kwargs.pop("extends", None)
            if extends and hasattr(extends, self.view_key):
                for k, v in getattr(extends, self.view_key).items():
                    kwargs.setdefault(k, v)

            kwargs.setdefault("route", rule)
            kwargs["decorators"] = kwargs.get("decorators", []) + f.decorators
            setattr(f, "_route_extends__", kwargs)
            setattr(f, "base_route", kwargs.get("route"))
            setattr(f, "decorators", kwargs.get("decorators", []))
        else:
            if not rule:
                raise ValueError("'rule' is missing in @route ")

            for k in _restricted_keys:
                if k in kwargs:
                    del kwargs[k]

            Shaft._bind_route_rule_cache(f, rule=rule, **kwargs)
        return f
    return decorator


def _accept_method(methods, f):
    kw = {
        "append_method": True,
        "methods": methods
    }
    Shaft._bind_route_rule_cache(f, rule=None, **kw)
    return f


def accept_get(f):
    """ Accept GET method """
    return _accept_method(["GET"], f)


def accept_post(f):
    """ Accept POST method """
    return _accept_method(["POST"], f)


def accept_post_get(f):
    """ Accept POST & GET method """
    return _accept_method(["POST",  "GET"], f)


def accept_delete(f):
    """ Accept DELETE method """
    return _accept_method(["DELETE"], f)


def accept_options(f):
    """ Accept OPTIONS method """
    return _accept_method(["OPTIONS"], f)


def accept_put(f):
    """ Accept PUT method """
    return _accept_method(["PUT"], f)


def template(page=None, layout=None, markup="html", **kwargs):
    """
    Decorator to change the view template and layout.

    It works on both Shaft class and view methods

    on class
        only $layout and markup are applied, everything else will be passed to the kwargs
        Using as first argument, it will be the layout.

        :first arg or $layout: The layout to use for that view
        :param layout: The layout to use for that view
        :param markup: the markup to use, by default it's html, can switch to jade.
        this will attach it to template_markup
        :param kwargs:
            get pass to the TEMPLATE_CONTEXT

    ** on method that return a dict
        page or layout are optional

        :param page: The html page
        :param layout: The layout to use for that view

        :param kwargs:
            get pass to the view as k/V

    ** on other methods that return other type, it doesn't apply

    :return:
    """
    pkey = "_template_extends__"

    def decorator(f):
        if inspect.isclass(f):
            layout_ = layout or page
            extends = kwargs.pop("extends", None)
            if extends and hasattr(extends, pkey):
                items = getattr(extends, pkey).items()
                if "layout" in items:
                    layout_ = items.pop("layout")
                for k, v in items:
                    kwargs.setdefault(k, v)
            if not layout_:
                layout_ = "layout.html"
            kwargs.setdefault("brand_name", "")
            kwargs["layout"] = layout_

            setattr(f, pkey, kwargs)
            setattr(f, "base_layout", kwargs.get("layout"))
            setattr(f, "template_markup", markup or "html")

            f.g(TEMPLATE_CONTEXT=kwargs)
            return f
        else:
            @functools.wraps(f)
            def wrap(*args2, **kwargs2):
                response = f(*args2, **kwargs2)
                if isinstance(response, dict) or response is None:
                    response = response or {}
                    if page:
                        response.setdefault("_template", page)
                    if layout:
                        response.setdefault("_layout", layout)
                    for k, v in kwargs.items():
                        response.setdefault(k, v)
                return response
            return wrap
    return decorator

# ------------------------------------------------------------------------------


def _normalize_response_tuple(tuple_):
    """
    Helper function to normalize view return values .
    It always returns (dict, status, headers). Missing values will be None.
    For example in such cases when tuple_ is
      (dict, status), (dict, headers), (dict, status, headers),
      (dict, headers, status)

    It assumes what status is int, so this construction will not work:
    (dict, None, headers) - it doesn't make sense because you just use
    (dict, headers) if you want to skip status.
    """
    v = tuple_ + (None,) * (3 - len(tuple_))
    return v if isinstance(v[1], int) else (v[0], v[2], v[1])


__view_parsers = set()
def view_parser(f):
    """
    A simple decorator to to parse the data that will be rendered
    :param func:
    :return:
    """
    __view_parsers.add(f)
    return f


def _build_response(data, renderer=None):
    """
    Build a response using the renderer from the data
    :return:
    """
    if isinstance(data, Response) or isinstance(data, BaseResponse):
        return data
    if not renderer:
        raise AttributeError(" Renderer is required")
    if isinstance(data, dict) or data is None:
        data = {} if data is None else data
        for _ in __view_parsers:
            data = _(data)
        return renderer(data), 200
    elif isinstance(data, tuple):
        data, status, headers = _normalize_response_tuple(data)
        for _ in __view_parsers:
            data = _(data)
        return renderer(data or {}), status, headers
    return data

json_renderer = lambda i, data: _build_response(data, jsonify)
xml_renderer = lambda i, data: _build_response(data, dicttoxml)


def render_json(func):
    """
    Decorator to render as JSON
    :param func:
    :return:
    """
    if inspect.isclass(func):
        apply_function_to_members(func, json)
        return func
    else:
        @functools.wraps(func)
        def decorated_view(*args, **kwargs):
            data = func(*args, **kwargs)
            return _build_response(data, jsonify)
        return decorated_view


def render_xml(func):
    """
    Decorator to render as XML
    :param func:
    :return:
    """
    if inspect.isclass(func):
        apply_function_to_members(func, xml)
        return func
    else:
        @functools.wraps(func)
        def decorated_view(*args, **kwargs):
            data = func(*args, **kwargs)
            return _build_response(data, dicttoxml)
        return decorated_view


def render_jsonp(func):
    """Wraps JSONified output for JSONP requests.
    http://flask.pocoo.org/snippets/79/
    """

    @functools.wraps(func)
    def decorated_view(*args, **kwargs):
        callback = request.args.get('callback', None)
        if callback:
            data = str(func(*args, **kwargs))
            content = str(callback) + '(' + data + ')'
            mimetype = 'application/javascript'
            return current_app.response_class(content, mimetype=mimetype)
        else:
            return func(*args, **kwargs)
    return decorated_view


# ------------------------------------------------------------------------------

class MenuNavigation(object):
    """
    MenuNavigation is class decorator to build page menu while building the enpoints

    Decorator to build navigation menu directly on the methods
    By default it will build the menu of the module, class an method
    If the class is also decorated, it will use the menu _name as the top level _name

    :param title: The menu title
    :param kwargs: extra options to pass into the menu or to move the menu somewhere else

        order int: The order of the menu in the set

        visible (list of bool or callback): To hide and show menu. Accepts bool or
                    list of callback function the callback function must return
                    a bool to check if all everything is True to show or will be False
                    ** When this menu is inside of a menu set, or has parent, if you want
                    that page to be activated, but don't want to create a menu link,
                    for example: a blog read page, set show to False. It will know
                    the menu set is active

        endpoint string: By default the endpoint is built based on the method and class.
                    When set it will be used instead

        endpoint_kwargs dict: dict of k/v data for enpoint

        group_name str: On class menu, it can be used to filter a menu set to display.
                    If a class is passed, it will try to inherit the group from that class

        The args below will allow you to change where the menu is placed.
        By default they are set automatically

        module_: the module _name. Usually if using another module
        class_: the class _name class _name in the module
        method_: The method _name, to build endpoint. Changing this will change the url

        extends: class Name. To use the extends of the class

        some other kwargs:
            url
            target
            fa_icon
            align_right
            show_profile_avatar
            show_profile_name
            css_class
            css_id
    :return:
    """

    _title_map = {}

    def __call__(self, title, **kwargs):

        def wrap(f):
            if title:
                module_ = kwargs.pop("module_", f.__module__)
                class_ = kwargs.pop("class_", inspect.stack()[1][3])
                method_ = kwargs.pop("method_", f.__name__)
                is_class_ = inspect.isclass(f)

                # if "visible_with_auth_user" in kwargs:
                #     kwargs["visible"] = ext.user_authenticated \
                #         if kwargs["visible_with_auth_user"] is True \
                #         else ext.user_not_authenticated

                extends = kwargs.pop("extends", None)
                if extends:
                    if is_class_:
                        extendscls = self.get(extends)
                        if extendscls and "kwargs" in extendscls:
                            if "group_name" not in kwargs:
                                kwargs["group_name"] = extendscls["kwargs"]["group_name"]
                            if "visible" not in kwargs:
                                kwargs["visible"] = extendscls["kwargs"]["visible"]
                    else:
                        if "endpoint" not in kwargs:
                            endpoint = "%s:%s" % (class_, f.__name__)
                            kwargs["endpoint"] = endpoint
                    module_ = extends.__module__
                    class_ = extends.__name__

                kwargs.setdefault("key", class_)

                self._push(module_=module_,
                           class_=class_,
                           method_=method_,
                           title=title,
                           is_class_=is_class_,
                           **kwargs)
            return f
        return wrap

    def __init__(self):
        self.MENU = {}

    def add(self, title, extends, **kwargs):
        """
        To manually add a extends. Usually, menu by hand which
        may not be attached to any functions
        :param title:
        :param cls:
        :param method_:
        :param is_class_:
        :param kwargs:
        :return:
        """
        f = extends
        self._push(title=title,
                   module_=f.__module__,
                   class_=f.__name__,
                   method_=f.__name__,
                   is_class_=False,
                   **kwargs)

    def clear(self):
        self.MENU = {}

    def _push(self, **kwargs):

        module_ = kwargs.pop("module_")
        class_ = kwargs.pop("class_")
        method_ = kwargs.pop("method_")
        is_class_ = kwargs.pop("is_class_")

        __cls = method_ if is_class_ else class_
        path = "%s.%s" % (module_, __cls)

        if path not in self.MENU:
            self.MENU[path] = {
                "title": None,
                "endpoint": None,
                "endpoint_kwargs": {},
                "order": None,
                "sub_menu": [],
                "kwargs": {}
            }

        if "title" not in kwargs:
            raise TypeError("Missing 'title' in menu decorator")

        _endpoint = "%s:%s" % (__cls,"index" if is_class_ else method_)
        kwargs["endpoint"] = kwargs.pop("endpoint", _endpoint)
        kwargs["endpoint_kwargs"] = kwargs.pop("endpoint_kwargs", {})
        order = kwargs.pop("order", 0)
        title = kwargs.pop("title")
        endpoint = kwargs.pop("endpoint")

        # visible: accepts a bool or list of callback to execute
        visible = kwargs.pop("visible", [True])
        if not isinstance(visible, list):
            visible = [visible]

        kwargs["visible"] = visible
        kwargs["active"] = None
        kwargs["index"] = None
        kwargs["key"] = __cls

        if is_class_:  # class menu
            kwargs["endpoint"] = endpoint
            kwargs["group_name"] = kwargs.pop("group_name", None)
            kwargs["has_submenu"] = True
            self.MENU[path]["title"] = title
            self.MENU[path]["order"] = order
            self.MENU[path]["kwargs"] = kwargs

        else:  # sub menu
            kwargs["has_submenu"] = False
            menu = (order, title, endpoint, kwargs)
            self.MENU[path]["sub_menu"].append(menu)

        self._title_map[endpoint] = title

    def _test_visibility(self, shows):
        if isinstance(shows, bool):
            return shows
        elif not isinstance(shows, list):
            shows = [shows]
        return all([x() if hasattr(x, "__call__") else x for x in shows])

    def get(self, cls):
        key = self.get_key(cls)
        return self.MENU[key]

    def get_key(self, cls):
        """
        Return the string key of the class
        :param cls: class
        :return: str
        """
        return "%s.%s" % (cls.__module__, cls.__name__)

    def get_extends(self, f):
        return dict(module_=f.__module__, class_=f.__name__)

    def render(self):
         # (order, Name, sub_menu, **kargs, hidden_submenu)
        menu_list = []
        menu_index = 0
        for _, menu in copy.deepcopy(self.MENU.items()):
            menu_index += 1
            sub_menu = []

            menu["kwargs"]["index"] = menu_index
            menu["kwargs"]["active"] = False
            if "visible" in menu["kwargs"]:
                menu["kwargs"]["visible"] = self._test_visibility(menu["kwargs"]["visible"])

            for s in menu["sub_menu"]:
                if s[2] == request.endpoint:
                    s[3]["active"] = True
                    menu["kwargs"]["active"] = True
                s[3]["visible"] = self._test_visibility(s[3]["visible"])
                sub_menu.append(s)

            _kwargs = menu["kwargs"]

            if menu["title"]:
                menu_list.append((
                    menu["order"],
                    menu["title"],
                    sorted(sub_menu),
                    _kwargs
                ))
            else:
                menu_list += sub_menu
        return sorted(menu_list)

    def init_app(self, app):

        def link_for(endpoint, props={}, **kwargs):
            url = url_for(endpoint, **kwargs)
            title = self._title_map.get(endpoint, "")
            props = " ".join(["%s='%s'" % (k, v) for k, v in props.items()])
            a = "<a href='{url}' {props}>{title}</a>".format(title=title,
                                                             url=url,
                                                             props=props)
            return Markup(a)

        @app.context_processor
        def _():
            return dict(link_for=link_for)

        @app.before_request
        def p(*args, **kwargs):
            if request.endpoint not in ["static", None]:
                # (order, Name, sub_menu, **kargs, hidden_submenu)
                Shaft.g(__NAV_MENU__=nav_menu.render())

nav_menu = MenuNavigation()
menu_title = nav_menu
init_app(nav_menu.init_app)


# ------------------------------------------------------------------------------

def cors(*args, **kwargs):
    """
    A wrapper around flask-cors cross_origin, to also act on classes

    **An extra note about cors, a response must be available before the
    cors is applied. Dynamic return is applied after the fact, so use the
    decorators, render_json, render_xml, or return self.render() for txt/html
    ie:
    @cors()
    class Index(Shaft):
        def index(self):
            return self.render()

        @render_json
        def json(self):
            return {}

    class Index2(Shaft):
        def index(self):
            return self.render()

        @cors()
        @render_json
        def json(self):
            return {}


    :return:
    """
    def decorator(fn):
        cors_fn = flask_cors.cross_origin(automatic_options=False, *args, **kwargs)
        if inspect.isclass(fn):
            apply_function_to_members(fn, cors_fn)
        else:
            return cors_fn(fn)
        return fn
    return decorator


def headers(params={}):
    """This decorator adds the headers passed in to the response
    http://flask.pocoo.org/snippets/100/
    """
    def decorator(f):

        if inspect.isclass(f):
            h = headers(params)
            apply_function_to_members(f, h)
            return f

        @functools.wraps(f)
        def decorated_function(*args, **kwargs):
            resp = make_response(f(*args, **kwargs))
            h = resp.headers
            for header, value in params.items():
                h[header] = value
            return resp
        return decorated_function
    return decorator


def noindex(f):
    """This decorator passes X-Robots-Tag: noindex
    http://flask.pocoo.org/snippets/100/
    """
    return headers({'X-Robots-Tag': 'noindex'})(f)


# ------------------------------------------------------------------------------
# Private signals
__signals_namespace = blinker.Namespace()

def emit_signal(sender=None, namespace=None):
    """
    @emit_signal
    A decorator to mark a method or function as a signal emitter
    :param sender: string  to be the sender.
    If empty, it will use the function __module__+__fn_name,
    or method __module__+__class_name__+__fn_name__
    :param namespace: The namespace. If None, it will use the global namespace
    :return:
    """
    if not namespace:
        namespace = __signals_namespace

    def decorator(fn):
        fname = sender
        if not fname:
            fnargs = inspect.getargspec(fn).args
            fname = fn.__module__
            if 'self' in fnargs or 'cls' in fnargs:
                caller = inspect.currentframe().f_back
                fname += "_" + caller.f_code.co_name
            fname += "__" + fn.__name__

        fn.pre = namespace.signal('pre_%s' % fname)
        fn.post = namespace.signal('post_%s' % fname)

        def send(action, *a, **kw):
            sig_name = "%s_%s" % (action, fname)
            result = kw.pop("result", None)
            kw.update(inspect.getcallargs(fn, *a, **kw))
            sendkw = {k: v for k, v in kw.items() if k in kw.keys()}
            sendkw["emitter"] = sendkw.pop('self', sendkw.pop('cls', kw.get('self', kw.get('cls', fn))))
            if action == 'post':
                sendkw["result"] = result
            namespace.signal(sig_name).send(fname, **sendkw)

        @functools.wraps(fn)
        def wrapper(*args, **kwargs):
            send('pre', *args, **kwargs)
            result = fn(*args, **kwargs)
            kwargs["result"] = result
            send('post', *args, **kwargs)
            return result
        return wrapper
    return decorator







"""
Custom Jinja filters
"""
import re
from jinja2 import Markup
from extras import md
from . import utils
from flask import url_for
import humanize
import arrow
from shaft import init_app, get_config


def format_datetime(dt, format):
    return "" if not dt else arrow.get(dt).format(format)


def format_date_time(dt):
    f = get_config("DATE_TIME_FORMAT", "MM/DD/YYYY h:mm a")
    return format_datetime(dt, f)


def format_date(dt):
    f = get_config("DATE_FORMAT", "MM/DD/YYYY")
    return format_datetime(dt, f)


def nl2br(s):
    """
    {{ s | nl2br }}

    Convert newlines into <p> and <br />s.
    """
    if not isinstance(s, basestring):
        s = str(s)
    s = re.sub(r'\r\n|\r|\n', '\n', s)
    paragraphs = re.split('\n{2,}', s)
    paragraphs = ['<p>%s</p>' % p.strip().replace('\n', '<br />') for p in
                  paragraphs]
    return '\n\n'.join(paragraphs)


def oembed(url, class_=""):
    """
    Create OEmbed link

    {{ url | oembed }}
    :param url:
    :param class_:
    :return:
    """
    o = "<a href=\"{url}\" class=\"oembed {class_}\" ></a>".format(url=url, class_=class_)
    return Markup(o)


def img_src(url, class_="", responsive=False, lazy_load=False, id_=""):
    """
    Create an image src

    {{ xyz.jpg | img_src }}

    :param url:
    :param class_:
    :param responsive:
    :param lazy_load:
    :param id_:
    :return:
    """
    if not url.startswith("http://") and not url.startswith("https://"):
        url = static_url(url)

    data_src = ""
    if responsive:
        class_ += " responsive"
    if lazy_load:
        data_src = url
        # 1x1 image
        url = "data:image/png;base64,iVBORw0KGgoAAAANSUhEUgAAAAEAAAABCAQAAAC1HAwCAAAAC0lEQVR42mNgYAAAAAMAASsJTYQAAAAASUVORK5CYII="
        class_ += " lazy"

    img = "<img src=\"{src}\" class=\"{class_}\" id=\"{id_}\" data-src={data_src}>"\
        .format(src=url, class_=class_, id_=id_, data_src=data_src)
    return Markup(img)


def static_url(url):
    """
    {{ url | static }}
    :param url:
    :return:
    """
    return url_for('static', filename=url)


FILTERS = {
    "slug": utils.slugify,  # slug
    "int_comma": humanize.intcomma,  # Transform an int to comma
    "strip_decimal": lambda d: d.split(".")[0],
    "bool_to_yes": lambda b: "Yes" if b else "No",
    "bool_to_int": lambda b: 1 if b else 0,
    "markdown": lambda text: Markup(md.html(text)),  # Return a markdown to HTML
    "markdown_toc": md.toc,  # Create a Table of Content of the Markdown
    "nl2br": nl2br,
    "format_datetime": format_datetime,  # Require the format to be passed
    "date": format_date,
    "date_time": format_date_time,
    "date_since": humanize.naturaldate,  # Show the date ago: Today, yesterday, July 27 (without year in same year), July 15 2014
    "time_since": humanize.naturaltime,  # To show the time ago: 3 min ago, 2 days ago, 1 year 7 days ago
    "oembed": oembed,
    "img_src": img_src,
    "static": static_url
}

def jinja_helpers(app):
    app.jinja_env.filters.update(FILTERS)

init_app(jinja_helpers)



"""
Shaft Bower

It's a dev tool for 3rd party assets using bower.

Its main purpose to only keep the necessary files of the 3rd party.

3rdparty vendors are place under /assets/Shaft/vendor

Do not put non bower files in the /assets/Shaft/vendor directory.

Place your files in /assets/Shaft/lib

Dependencies: Bower

To run this file: python shaft-bower.py

"""

import os
import sh
import json
from shaft import utils

VERSION = "0.1.0"

DEPENDENCIES = {
    "jquery": {
        "version": "~2.2.4",
        "copy": ["dist/jquery.min.js"]
    },
    "bootstrap": {
        "version": "~3.3",
        "copy": ["dist/css",
                 "dist/fonts",
                 "dist/js"]
    },
    "font-awesome": {
        "version": "~4.6",
        "copy": ["css/", "fonts/"]
    },
    "jquery-form-validator": {
        "version": "~2.2.8",
        "copy": ["form-validator/"]
    },
    "jquery-lazy": {
        "version": "~1.6.5",
        "copy": ["jquery.lazy.min.js"]
    },
    "js-cookie": {
        "version": "~2.1.0",
        "copy": ["src/js.cookie.js"]
    },
    "jssocials": {
        "version": "~1.1.0",
        "copy": ["dist/jssocials.js",
                 "dist/jssocials.css",
                 "dist/jssocials-theme-flat.css"]
    },
    "bootstrap-social": {
        "version": "~4.9.0",
        "copy": ["bootstrap-social.css"]
    },
    "eonasdan-bootstrap-datetimepicker": {
        "version": "~4.17.37",
        "copy": ["build/css/bootstrap-datetimepicker.min.css",
                 "build/js/bootstrap-datetimepicker.min.js"]
    },
    "moment": {
        "version": "~2.8",
        "copy":["moment.js"]
    },
    "moment-timezone": {
        "version": "~0.4",
        "copy": ["moment-timezone.js",
                 "moment-timezone-utils.js"]
    }

}

# ------------------------------------------------------------------------------
# DO NOT CHANGE BELOW
# ------------------------------------------------------------------------------

CWD = os.getcwd()
VENDOR_DIR = "%s/Shaft/vendor" % CWD
BOWER_COMPONENTS_DIR = "%s/bower_components" % CWD

print("-" * 80)
print("Shaft Bower")
print("")

bower = {
    "name": "Shaft",
    "version": VERSION,
    "dependencies": {d:v["version"] for d, v in DEPENDENCIES.items()}
}

if os.path.exists(VENDOR_DIR):
    utils.remove_dir(VENDOR_DIR)

if not os.path.exists(VENDOR_DIR):
    utils.make_dirs(VENDOR_DIR)

print("Saving bower.json ...")
with open("./bower.json", "w+") as f:
    f.write(json.dumps(bower))

print("Running bower install ...")
with sh.pushd(CWD):
    #sh.bower("cache", "clean")
    sh.bower("install", "--force-latest")
    #sh.bower("prune")

print("Copying files to vendor directories ...")
for name, v in DEPENDENCIES.items():
    vendor_dir = "%s/%s" % (VENDOR_DIR, name)
    if not os.path.exists(vendor_dir):
        utils.make_dirs(vendor_dir)
    if "copy" in v:
        for f in v["copy"]:
            source = "%s/%s/%s" % (BOWER_COMPONENTS_DIR, name, f)
            dest = "%s/%s" % (vendor_dir, f)
            f = f.rsplit("/", 1)[1] if "/" in f else f
            dest = "%s/%s" % (vendor_dir, f)
            if f.endswith(".css") or f.endswith(".js"):
                utils.copy_file(source, dest)
            else:
                utils.copy_dir(source, dest)
                pass

#-----

# FONT AWESOME
# Cleanup font awesome to make sure it points to the right fonts
print("Copying Font-Awesome...")
fa_src = "%s/font-awesome" % VENDOR_DIR
for _ in ["font-awesome.css", "font-awesome.min.css"]:
    with open("%s/%s" % (fa_src, _), "r+") as f:
        content = f.read().replace("../fonts/", "./")
        f.seek(0)
        f.write(content)
        f.truncate()

# Copy Font-awesome to the skeleton.
# It is needed since it will be loaded from static
fa_dest = "../../../skel/assets/static/vendor/font-awesome"
if not os.path.exists(fa_dest):
    utils.make_dirs(fa_dest)
utils.copy_dir(fa_src, fa_dest)


print("Cleaning up...")
if os.path.exists(BOWER_COMPONENTS_DIR):
    utils.remove_dir(BOWER_COMPONENTS_DIR)

print("Done!")





class AuthError(Exception): pass
class LoginError(AuthError): pass
class SecureLoginError(AuthError): pass
class InvalidLoginError(LoginError): pass
class InvalidLoginPassword(LoginError): pass
class VerifyEmailError(LoginError): pass



"""
Shaft

"""

import re
import os
import sys
import inspect
import datetime
import functools
import logging
import logging.config
import pkg_resources
import jinja2
import active_alchemy
from six import string_types
from importlib import import_module
from werkzeug import import_string
from flask_assets import Environment
from sqlalchemy.engine.url import make_url as sa_make_url
from werkzeug.contrib.fixers import ProxyFix
from werkzeug.routing import (BaseConverter,
                              parse_rule)
from flask import (Flask,
                   g,
                   render_template,
                   flash,
                   session,
                   make_response,
                   Response,
                   request,
                   abort,
                   url_for as f_url_for,
                   redirect as f_redirect)
from . import (utils,
               exceptions)
from .__about__ import *

_py2 = sys.version_info[0] == 2

# ------------------------------------------------------------------------------

__all__ = [
            "Shaft",
            "ShaftInit",
            "db",
            "models",
            "views",
            "get_env",
            "set_env",
            "get_app_env",
            "get_env_config",
            "get_config",
            "page_meta",
            "flash_success",
            "flash_error",
            "flash_info",
            "flash_data",
            "get_flash_data",
            "init_app",
            "register_package",
            "register_models",
            "abort",
            "redirect",
            "url_for",

            # For convenience when importing from shaft, but can use
            # the flask one
            "flash",
            "session",
            "request",

            # Not sure why it's there
            "import_module",
           ]


# Will hold all active class views
# It can be used for redirection etc
# ie: redirect(views.ContactPage.index)
views = type('', (), {})

# Will hold models from apps, or to be shared
# ie, set new model property -> models.MyNewModel = MyModel
# ie: use property -> models.MyNewModel.all()
# For convenience, use `register_models(**kw)` to register the models
# By default shaft will load all the application/models.py models
models = type('', (), {})


def register_models(**kwargs):
    """
    Alias to register model
    :param kwargs:
    :return:
    """
    [setattr(models, k, v) for k, v in kwargs.items()]


# Hold the current environment
__ENV__ = None


def set_env(env):
    """
    Set the envrionment manually
    :param env:
    :return:
    """
    global __ENV__
    __ENV__ = env.lower().capitalize()


def get_env():
    """
    Return the Capitalize environment name
    It can be used to retrieve class base config
    Default: Development
    :returns: str Capitalized
    """
    if not __ENV__:
        env = os.environ["env"] if "env" in os.environ else "Development"
        set_env(env)
    return __ENV__


def get_app_env():
    """
    if the app and the envi are passed in the command line as 'app=app:env'
    :return: tuple app, env
    """
    app, env = None, get_env()
    if "app" in os.environ:
        app = os.environ["app"].lower()
        if ":" in app:
            app, env = os.environ["app"].split(":", 2)
            set_env(env)
    return app, env


def get_env_config(config):
    """
    Return config class based based on the config
    :param config : Object - The configuration module containing the environment object
    """
    return getattr(config, get_env())


def init_app(kls):
    """
    To bind middlewares, plugins that needs the 'app' object to init
    Bound middlewares will be assigned on cls.init()
    """
    if not hasattr(kls, "__call__"):
        raise exceptions.ShaftError("init_app: '%s' is not callable" % kls)
    Shaft._init_apps.add(kls)
    return kls


def register_package(pkg):
    """
    Allow to register an app packages by loading and exposing: templates, static,
    and exceptions for abort()

    Structure of package
        root
            | $package_name
                | __init__.py
                |
                | /templates
                    |
                    |
                |
                | /static
                    |
                    | assets.yml

    :param pkg: str - __package__
                    or __name__
                    or The root dir
                    or the dotted resource package (package.path.path,
                    usually __name__ of templates and static
    """

    root_pkg_dir = pkg
    if not os.path.isdir(pkg) and "." in pkg:
        root_pkg_dir = pkg_resources.resource_filename(pkg, "")

    template_path = os.path.join(root_pkg_dir, "templates")
    static_path = os.path.join(root_pkg_dir, "static")

    logging.info("Registering App: " + pkg)
    if os.path.isdir(template_path):
        template_path = jinja2.FileSystemLoader(template_path)
        Shaft._template_paths.add(template_path)

    if os.path.isdir(static_path):
        Shaft._static_paths.add(static_path)
        Shaft._add_asset_bundle(static_path)


def get_config(key, default=None):
    """
    Shortcut to access the application's config in your class
    :param key: The key to access
    :param default: The default value when None
    :returns mixed:
    """
    return Shaft._app.config.get(key, default)


def page_meta(**kwargs):
    """
    Meta allows you to add page meta data
    :params **kwargs:

    meta keys we're expecting:
        title (str)
        description (str)
        url (str) (Will pick it up by itself if not set)
        image (str)
        site_name (str) (but can pick it up from config file)
        object_type (str)
        keywords (list)
        locale (str)
        card (str)

        **Boolean By default these keys are True
        use_opengraph
        use_twitter
        use_googleplus
python
    """
    meta = Shaft._global.get("__META__", {})
    meta.update(**kwargs)
    Shaft.g(__META__=meta)


def flash_success(msg):
    """
    Alias to flash, but set a success message
    :param msg:
    :return:
    """
    return flash(msg, "success")


def flash_error(msg):
    """
    Alias to flash, but set an error message
    :param msg:
    :return:
    """
    return flash(msg, "error")


def flash_info(msg):
    """
    Alias to flash, but set an info message
    :param msg:
    :return:
    """
    return flash(msg, "info")


def flash_data(data):
    """
    Just like flash, but will save data
    :param data:
    :return:
    """
    session["_flash_data"] = data


def get_flash_data():
    """
    Retrieved
    :return: mixed
    """
    return session.pop("_flash_data", None)


is_method = lambda x: inspect.ismethod if _py2 else inspect.isfunction


# ------------------------------------------------------------------------------

def _get_action_endpoint(action):
    """
    Return the endpoint base on the view's action
    :param action:
    :return:
    """
    _endpoint = None
    if is_method(action):
        if hasattr(action, "_rule_cache"):
            rc = action._rule_cache
            if rc:
                k = list(rc.keys())[0]
                rules = rc[k]
                len_rules = len(rules)
                if len_rules == 1:
                    rc_kw = rules[0][1]
                    _endpoint = rc_kw.get("endpoint", None)
                    if not _endpoint:
                        _endpoint = _build_endpoint_route_name(action)
                elif len_rules > 1:
                    _prefix = _build_endpoint_route_name(action)
                    for r in Shaft._app.url_map.iter_rules():
                        if ('GET' in r.methods or 'POST' in r.methods) \
                                and _prefix in r.endpoint:
                            _endpoint = r.endpoint
                            break
    return _endpoint


def url_for(endpoint, **kw):
    """
    Shaft url_for is an alias to the flask url_for, with the ability of
    passing the function signature to build the url, without knowing the endpoint
    :param endpoint:
    :param kw:
    :return:
    """

    _endpoint = None
    if isinstance(endpoint, string_types):
        return f_url_for(endpoint, **kw)
    else:
        # self, will refer the caller method, by getting the method name
        if isinstance(endpoint, Shaft):
            fn = sys._getframe().f_back.f_code.co_name
            endpoint = getattr(endpoint, fn)

        if is_method(endpoint):
            _endpoint = _get_action_endpoint(endpoint)
            if not _endpoint:
                _endpoint = _build_endpoint_route_name(endpoint)
    if _endpoint:
        return f_url_for(_endpoint, **kw)
    else:
        raise exceptions.ShaftError('Shaft `url_for` received an invalid endpoint')


def redirect(endpoint, **kw):
    """
    Redirect allow to redirect dynamically using the classes methods without
    knowing the right endpoint.
    Expecting all endpoint have GET as method, it will try to pick the first
    match, based on the endpoint provided or the based on the Rule map_url

    An endpoint can also be passed along with **kw

    An http: or https: can also be passed, and will redirect to that site.

    example:
        redirect(self.hello_world)
        redirect(self.other_page, name="x", value="v")
        redirect("https://google.com")
        redirect(views.ContactPage.index)
    :param endpoint:
    :return: redirect url
    """

    _endpoint = None

    if isinstance(endpoint, string_types):
        _endpoint = endpoint
        # valid for https:// or /path/
        # Endpoint should not have slashes. Use : (colon) to build endpoint
        if "/" in endpoint:
            return f_redirect(endpoint)
        else:
            for r in Shaft._app.url_map.iter_rules():
                _endpoint = endpoint
                if 'GET' in r.methods and endpoint in r.endpoint:
                    _endpoint = r.endpoint
                    break
    else:
        # self, will refer the caller method, by getting the method name
        if isinstance(endpoint, Shaft):
            fn = sys._getframe().f_back.f_code.co_name
            endpoint = getattr(endpoint, fn)

        if is_method(endpoint):
            _endpoint = _get_action_endpoint(endpoint)
            if not _endpoint:
                _endpoint = _build_endpoint_route_name(endpoint)
    if _endpoint:
        return f_redirect(url_for(_endpoint, **kw))
    else:
        raise exceptions.ShaftError("Invalid endpoint")


def _build_endpoint_route_name(endpoint):

    cls = endpoint.im_class() \
        if (not hasattr(endpoint, "__self__") or endpoint.__self__ is None) \
        else endpoint.__self__

    return cls.build_route_name(endpoint.__name__)


# ------------------------------------------------------------------------------

class _ShaftAlchemy(active_alchemy.ActiveAlchemy):
    """
    A custom ActiveAlchemy wrapper which defers the connection
    """
    def __init__(self):
        self.Model = active_alchemy.declarative_base(cls=active_alchemy.Model, name='Model')
        self.BaseModel = active_alchemy.declarative_base(cls=active_alchemy.BaseModel, name='BaseModel')

    def _connect(self, uri, app):
        self.uri = uri
        self.info = sa_make_url(uri)
        self.options = self._cleanup_options(
            echo=False,
            pool_size=None,
            pool_timeout=None,
            pool_recycle=None,
            convert_unicode=True,
        )

        self.connector = None
        self._engine_lock = active_alchemy.threading.Lock()
        self.session = active_alchemy._create_scoped_session(self,
                                                             query_cls=active_alchemy.BaseQuery)

        self.Model.db, self.BaseModel.db = self, self
        self.Model._query, self.BaseModel._query = self.session.query, self.session.query

        self.init_app(app)
        active_alchemy._include_sqlalchemy(self)

# Setup the DB
# upon initialization will use the right URI for it
# also, it exposes the db object to all modules
db = _ShaftAlchemy()


class Shaft(object):
    """Base view for any class based views implemented with Flask-Classy. Will
    automatically configure routes when registered with a Flask app instance.
    Credit: Shout out to Flask-Classy for the greatest logic in this class
    Flask-Classy -> https://github.com/apiguy/flask-classy
    """

    decorators = []
    base_route = None
    route_prefix = None
    trailing_slash = True
    base_layout = "layout.html"
    template_markup = "html"
    assets = None
    logger = None
    _ext = set()
    __special_methods = ["get", "put", "patch", "post", "delete", "index"]
    _installed_apps = []
    _app = None
    _init_apps = set()
    _template_paths = set()
    _static_paths = set()
    _asset_bundles = set()
    _default_page_meta = dict(
            title="",
            description="",
            url="",
            image="",
            site_name="",
            object_type="article",
            locale="",
            keywords=[],
            use_opengraph=True,
            use_googleplus=True,
            use_twitter=True,
            properties={}
        )
    _global = dict(
        __NAME__=__title__,
        __VERSION__=__version__,
        __YEAR__=datetime.datetime.now().year,
        __META__=_default_page_meta
    )

    @classmethod
    def __call__(cls,
                 flask_or_import_name,
                 app_name=None,
                 app_directory="apps",
                 implicit_import=True,
                 load_app=True,
                 load_installed_apps=True
                 ):
        """
        Allow to register all subclasses of Shaft at once

        If a class doesn't have a route base, it will create a dasherize version
        of the class name.

        So we call it once initiating
        :param flask_or_import_name: Flask instance or import name -> __name__
        :param app_name: name of the project. If the directory and config is empty, it will guess them from here
        :param app_directory: the directory name relative to the current execution path
        :param implicit_import: when True, it will automatically load the views, models
        :param load_app: when True it will load your application
        :param load_installed_apps: bool - When True it will load the installed apps
        """

        app = flask_or_import_name \
            if isinstance(flask_or_import_name, Flask) \
            else Flask(flask_or_import_name)

        app_name = app_name if app_name else get_app_env()[0] or "www"

        app_env = get_env()

        # Load global and project config
        for c in [
                  "%s.config.%s" % (app_directory, app_env),  # Global config
                  "%s.%s.config.%s" % (app_directory, app_name, app_env)  # project config
                  ]:
            try:
                app.config.from_object(c)
            except ImportError as e:
                pass

        app.wsgi_app = ProxyFix(app.wsgi_app)
        app.url_map.converters['regex'] = RegexConverter
        app.template_folder = "%s/%s/templates" % (app_directory, app_name)
        app.static_folder = "%s/%s/static" % (app_directory, app_name)

        cls._app = app

        cls._load_extensions()

        cls._setup_logger()

        # Flask Assets
        cls.assets = Environment(cls._app)

        cls._setup_db()

        # INSTALLED_APPS
        if load_installed_apps:
            _ = cls.setup_installed_apps()

        cls._expose_models()

        # Implicit import.
        # Shaft can load models and view files according to the convention:
        # application.models,
        # application.$project_name.models,
        # application.$project_name.views
        if implicit_import:
            # Import global application's models and project's models
            for m in ["%s.models" % app_directory,
                      "%s.%s.models" % (app_directory, app_name)
                      ]:
                try:
                    import_module(m)
                    cls._expose_models()
                except ImportError as ie1:
                    pass

            # The view is required
            import_module("%s.%s.views" % (app_directory, app_name))

        cls._expose_models()

        # Setup init_app
        # init_app instanciate functions that may need the flask.app object
        # Usually for flask extension to be setup
        [_app(cls._app) for _app in cls._init_apps]

        if load_app:
            # Ad bundles
            cls._add_asset_bundle(cls._app.static_folder)

            # Register templates
            if cls._template_paths:
                loader = [cls._app.jinja_loader] + list(cls._template_paths)
                cls._app.jinja_loader = jinja2.ChoiceLoader(loader)
                
            # Static
            if cls._static_paths:
                cls.assets.load_path = [cls._app.static_folder] + list(cls._static_paths)
                [cls.assets.from_yaml(a) for a in cls._asset_bundles]
                
            # Register views
            for subcls in cls.__subclasses__():
                base_route = subcls.base_route
                if not base_route:
                    base_route = utils.dasherize(utils.underscore(subcls.__name__))
                    if subcls.__name__.lower() == "index":
                        base_route = "/"
                subcls._register(cls._app, base_route=base_route)

        @cls._app.after_request
        def _after_request_cleanup(response):
            cls._global["__META__"] = cls._default_page_meta.copy()
            return response
        return cls._app

    @classmethod
    def setup_installed_apps(cls):
        """
        To import 3rd party applications along with associated properties

        It is a list of dict or string.

        When a dict, it contains the `app` key and the configuration,
        if it's a string, it is just the app name

        If you require dependencies from other packages, dependencies
        must be placed before the calling package.

        It is required that __init__ in the package app has an entry point method
        -> 'main(**kw)' which will be used to setup the default app.

        By default the following **kwargs are passed to the main function:
            app: (object)[always] object of the flask app
            route: (string) the base route for the app. default: /
            nav_menu: (dict) the navigation config for navigation menu. By default the title is set to None
            decorators: (list) list of decorators (as string) to use
            options: (dict) all the app's customs options

        As a dict
        INSTALLED_APPS = [
            "my.contrib.app",
            "my.other.contrib.app",
            {
                "app": "my.contrib.app3",
                "options": {

                },
                "modules": {
                    "module_a": {
                        "route": "/",
                        "nav_menu": {

                        }
                    },
                    "module_b": {
                        "route": "/",
                        "decorators": []
                    }
                }
            }
        ]
        :return:
        """
        cls._installed_apps = cls._app.config.get("INSTALLED_APPS", [])
        if cls._installed_apps:
            _ = []

            def import_app(app, kwargs):
                try:
                    _.append(import_string("%s.main" % app)(**kwargs))
                except ImportError as e:
                    raise e
                    #logging.warning(e.message)

            for k in cls._installed_apps:
                if isinstance(k, string_types):
                    a = k
                    prop = {}
                elif isinstance(k, dict):
                    a = k.get("app")
                    k["name"] = a
                    prop = k

                _kwargs = {
                    "app": cls._app,
                    "name": prop.get("name"),
                    "route": prop.get("route") or "/",
                    "nav_menu": prop.get("nav_menu") or {"title": None},
                    "decorators": [import_string(d) for d in prop.get("decorators")]
                                    if "decorators" in prop
                                    else [],
                    "options": prop.get("options") or {}
                }

                # Loading the module
                try:
                    _impm = import_string(a)
                    #logging.info("Loading installed app: %s v.%s" % (a, _impm.__version__))
                    try:
                        import_string("%s.models" % a)
                        cls._expose_models()
                    except ImportError as ie: pass
                    import_app(a, _kwargs)
                except ImportError as e:
                    #logging.warning(e)
                    raise e

                modules = prop["modules"] if "modules" in prop else None
                if modules:
                    if isinstance(modules, list):
                        for m in modules:
                            import_app("%s.%s" % (a, m), _kwargs)
                    elif isinstance(modules, dict):
                        for n, kw in modules.items():
                            kw2 = {
                                "app": _kwargs["app"],
                                "name": "%s.%s" % (a, n),
                                "route": kw.get("route") or "/",
                                "nav_menu": kw.get("nav_menu") or {"title": None},
                                "decorators": _kwargs["decorators"],
                                "options": _kwargs["options"]
                            }
                            import_app("%s.%s" % (a, n), kw2)
            return _

    @classmethod
    def render(cls, data={}, _template=None, _layout=None, **kwargs):
        """
        Render the view template based on the class and the method being invoked
        :param data: The context data to pass to the template
        :param _template: The file template to use. By default it will map the classname/action.html
        :param _layout: The body layout, must contain {% include __template__ %}
        """
        if not _template:
            stack = inspect.stack()[1]
            module = inspect.getmodule(cls).__name__
            module_name = module.split(".")[-1]
            action_name = stack[3]      # The method being called in the class
            view_name = cls.__name__    # The name of the class without View

            if view_name.endswith("View"):
                view_name = view_name[:-4]
            _template = "%s/%s.%s" % (view_name, action_name, cls.template_markup)

        data = data or dict()
        if kwargs:
            data.update(kwargs)
        data["__"] = cls._global
        data["__template__"] = _template

        return render_template(_layout or cls.base_layout, **data)

    @classmethod
    def g(cls, **kwargs):
        """
        Assign a global view context to be used in the template
        :params **kwargs:
        """
        cls._global.update(kwargs)

    @classmethod
    def _add_asset_bundle(cls, path):
        """
        Add a webassets bundle yml file
        """
        f = "%s/assets.yml" % path
        if os.path.isfile(f):
            cls._asset_bundles.add(f)

    @classmethod
    def _setup_logger(cls):
        logging_config = cls._app.config.get("LOGGING")
        if not logging_config:
            logging_config = {
                "version": 1,
                "handlers": {
                    "default": {
                        "class": cls._app.config.get("LOGGING_CLASS", "logging.StreamHandler")
                    }
                },
                'loggers': {
                    '': {
                        'handlers': ['default'],
                        'level': 'WARN',
                    }
                }
            }

        logging.config.dictConfig(logging_config)
        cls.logger = logging.getLogger("root")
        cls._app._logger = cls.logger
        cls._app._loger_name = cls.logger.name

    @classmethod
    def _setup_db(cls):
        cls._app.db = None
        uri = cls._app.config.get("DB_URI")
        if uri:
            db._connect(uri, cls._app)
            cls._app.db = db

    @classmethod
    def _expose_models(cls):
        if cls._app.db:
            register_models(**{m.__name__:m
                               for m in cls._app.db.Model.__subclasses__()
                               if not hasattr(models, m.__name__)})

    @classmethod
    def _register(cls, app, base_route=None, subdomain=None, route_prefix=None,
                 trailing_slash=None):
        """Registers a Shaft class for use with a specific instance of a
        Flask app. Any methods not prefixes with an underscore are candidates
        to be routed and will have routes registered when this method is
        called.

        :param app: an instance of a Flask application

        :param base_route: The base path to use for all routes registered for
                           this class. Overrides the base_route attribute if
                           it has been set.

        :param subdomain:  A subdomain that this registration should use when
                           configuring routes.

        :param route_prefix: A prefix to be applied to all routes registered
                             for this class. Precedes base_route. Overrides
                             the class' route_prefix if it has been set.
        """

        if cls is Shaft:
            raise TypeError("cls must be a subclass of Shaft, not Shaft itself")

        # Assign views
        setattr(views, cls.__name__, cls)

        if base_route:
            cls.orig_base_route = cls.base_route
            cls.base_route = base_route

        if route_prefix:
            cls.orig_route_prefix = cls.route_prefix
            cls.route_prefix = route_prefix

        if not subdomain:
            if hasattr(app, "subdomain") and app.subdomain is not None:
                subdomain = app.subdomain
            elif hasattr(cls, "subdomain"):
                subdomain = cls.subdomain

        if trailing_slash is not None:
            cls.orig_trailing_slash = cls.trailing_slash
            cls.trailing_slash = trailing_slash

        for name, value in get_interesting_members(Shaft, cls):
            proxy = cls.make_proxy_method(name)
            route_name = cls.build_route_name(name)
            try:
                if hasattr(value, "_rule_cache") and name in value._rule_cache:
                    for idx, cached_rule in enumerate(value._rule_cache[name]):
                        rule, options = cached_rule
                        rule = cls.build_rule(rule)
                        sub, ep, options = cls.parse_options(options)

                        if not subdomain and sub:
                            subdomain = sub

                        if ep:
                            endpoint = ep
                        elif len(value._rule_cache[name]) == 1:
                            endpoint = route_name
                        else:
                            endpoint = "%s_%d" % (route_name, idx,)

                        app.add_url_rule(rule, endpoint, proxy,
                                         subdomain=subdomain,
                                         **options)
                elif name in cls.__special_methods:
                    if name in ["get", "index"]:
                        methods = ["GET"]
                        if name == "index":
                            if hasattr(value, "_methods_cache"):
                                methods = value._methods_cache
                    else:
                        methods = [name.upper()]

                    rule = cls.build_rule("/", value)
                    if not cls.trailing_slash:
                        rule = rule.rstrip("/")
                    app.add_url_rule(rule, route_name, proxy,
                                     methods=methods,
                                     subdomain=subdomain)

                else:
                    methods = value._methods_cache \
                        if hasattr(value, "_methods_cache") \
                        else ["GET"]

                    name = utils.dasherize(name)
                    route_str = '/%s/' % name
                    if not cls.trailing_slash:
                        route_str = route_str.rstrip('/')
                    rule = cls.build_rule(route_str, value)
                    app.add_url_rule(rule, route_name, proxy,
                                     subdomain=subdomain,
                                     methods=methods)
            except DecoratorCompatibilityError:
                raise DecoratorCompatibilityError("Incompatible decorator detected on %s in class %s" % (name, cls.__name__))

        if hasattr(cls, "orig_base_route"):
            cls.base_route = cls.orig_base_route
            del cls.orig_base_route

        if hasattr(cls, "orig_route_prefix"):
            cls.route_prefix = cls.orig_route_prefix
            del cls.orig_route_prefix

        if hasattr(cls, "orig_trailing_slash"):
            cls.trailing_slash = cls.orig_trailing_slash
            del cls.orig_trailing_slash

    @classmethod
    def parse_options(cls, options):
        """Extracts subdomain and endpoint values from the options dict and returns
           them along with a new dict without those values.
        """
        options = options.copy()
        subdomain = options.pop('subdomain', None)
        endpoint = options.pop('endpoint', None)
        return subdomain, endpoint, options,

    @classmethod
    def make_proxy_method(cls, name):
        """Creates a proxy function that can be used by Flasks routing. The
        proxy instantiates the Shaft subclass and calls the appropriate
        method.
        :param name: the name of the method to create a proxy for
        """

        i = cls()
        view = getattr(i, name)

        for decorator in cls.decorators:
            view = decorator(view)

        @functools.wraps(view)
        def proxy(**forgettable_view_args):
            # Always use the global request object's view_args, because they
            # can be modified by intervening function before an endpoint or
            # wrapper gets called. This matches Flask's behavior.
            del forgettable_view_args

            if hasattr(i, "before_request"):
                response = i.before_request(name, **request.view_args)
                if response is not None:
                    return response

            before_view_name = "before_" + name
            if hasattr(i, before_view_name):
                before_view = getattr(i, before_view_name)
                response = before_view(**request.view_args)
                if response is not None:
                    return response

            response = view(**request.view_args)

            # You can also return a dict or None, it will pass it to render
            if isinstance(response, dict) or response is None:
                response = response or {}
                if hasattr(i, "_renderer"):
                    response = i._renderer(response)
                else:
                    df_v_t = "%s/%s.%s" % (cls.__name__, view.__name__, cls.template_markup)
                    response.setdefault("_template", df_v_t)
                    response = i.render(**response)

            if not isinstance(response, Response):
                response = make_response(response)

            for ext in cls._ext:
                response = ext(response)

            after_view_name = "after_" + name
            if hasattr(i, after_view_name):
                after_view = getattr(i, after_view_name)
                response = after_view(response)

            if hasattr(i, "after_request"):
                response = i.after_request(name, response)

            return response

        return proxy

    @classmethod
    def build_rule(cls, rule, method=None):
        """Creates a routing rule based on either the class name (minus the
        'View' suffix) or the defined `base_route` attribute of the class

        :param rule: the path portion that should be appended to the
                     route base

        :param method: if a method's arguments should be considered when
                       constructing the rule, provide a reference to the
                       method here. arguments named "self" will be ignored
        """

        rule_parts = []

        if cls.route_prefix:
            rule_parts.append(cls.route_prefix)

        base_route = cls.get_base_route()
        if base_route:
            rule_parts.append(base_route)

        rule_parts.append(rule)
        ignored_rule_args = ['self']
        if hasattr(cls, 'base_args'):
            ignored_rule_args += cls.base_args

        if method:
            args = get_true_argspec(method)[0]
            for arg in args:
                if arg not in ignored_rule_args:
                    rule_parts.append("<%s>" % arg)

        result = "/%s" % "/".join(rule_parts)
        return re.sub(r'(/)\1+', r'\1', result)

    @classmethod
    def get_base_route(cls):
        """Returns the route base to use for the current class."""

        base_route = cls.__name__.lower()
        if cls.base_route is not None:
            base_route = cls.base_route
            base_rule = parse_rule(base_route)
            cls.base_args = [r[2] for r in base_rule]

        return base_route.strip("/")

    @classmethod
    def build_route_name(cls, method_name):
        """Creates a unique route name based on the combination of the class
        name with the method name.

        :param method_name: the method name to use when building a route name
        """
        return cls.__name__ + ":%s" % method_name

    @staticmethod
    def _bind_route_rule_cache(f, rule, append_method=False, **kwargs):
        # Put the rule cache on the method itself instead of globally
        if rule is None:
            rule = utils.dasherize(f.__name__) + "/"
        if not hasattr(f, '_rule_cache') or f._rule_cache is None:
            f._rule_cache = {f.__name__: [(rule, kwargs)]}
        elif not f.__name__ in f._rule_cache:
            f._rule_cache[f.__name__] = [(rule, kwargs)]
        else:
            # when and endpoint accepts multiple METHODS, ie: post(), get()
            if append_method:
                for r in f._rule_cache[f.__name__]:
                    if r[0] == rule and "methods" in r[1] and "methods" in kwargs:
                        r[1]["methods"] = list(set(r[1]["methods"] + kwargs["methods"]))
            else:
                f._rule_cache[f.__name__].append((rule, kwargs))
        return f

    @classmethod
    def _load_extensions(cls):
        extensions = [
            'pyjade.ext.jinja.PyJadeExtension',
            'shaft.extras.jade.JadeTagExtension',
            'shaft.extras.md.MarkdownExtension',
            'shaft.extras.md.MarkdownTagExtension',
        ]
        if cls._app.config.get("COMPRESS_HTML"):
            extensions.append('shaft.extras.htmlcompress.HTMLCompress')
        for ext in extensions:
            cls._app.jinja_env.add_extension(ext)



# ShaftInit
ShaftInit = Shaft()

# ------------------------------------------------------------------------------


def get_interesting_members(base_class, cls):
    """Returns a generator of methods that can be routed to"""

    base_members = dir(base_class)
    predicate = inspect.ismethod if _py2 else inspect.isfunction
    all_members = inspect.getmembers(cls, predicate=predicate)
    return (member for member in all_members
            if not member[0] in base_members
            and ((hasattr(member[1], "__self__") and not member[1].__self__ in inspect.getmro(cls)) if _py2 else True)
            and not member[0].startswith("_")
            and not member[0].startswith("before_")
            and not member[0].startswith("after_"))


def apply_function_to_members(cls, fn):
    for name, method in get_interesting_members(Shaft, cls):
        setattr(cls, name, fn(method))


def get_true_argspec(method):
    """Drills through layers of decorators attempting to locate the actual argspec for the method."""

    argspec = inspect.getargspec(method)
    args = argspec[0]
    if args and args[0] == 'self':
        return argspec
    if hasattr(method, '__func__'):
        method = method.__func__
    if not hasattr(method, '__closure__') or method.__closure__ is None:
        raise DecoratorCompatibilityError

    closure = method.__closure__
    for cell in closure:
        inner_method = cell.cell_contents
        if inner_method is method:
            continue
        if not inspect.isfunction(inner_method) \
            and not inspect.ismethod(inner_method):
            continue
        true_argspec = get_true_argspec(inner_method)
        if true_argspec:
            return true_argspec


class DecoratorCompatibilityError(Exception):
    pass


class RegexConverter(BaseConverter):
    def __init__(self, url_map, *items):
        super(RegexConverter, self).__init__(url_map)
        self.regex = items[0]


def get_installed_app_options(app):
    """
    Get the app's options
    :param app: the app, ie: __name__
    :return:
    """
    options = {}
    if Shaft._installed_apps:
        for k in Shaft._installed_apps:
            if isinstance(k, dict) and "name" in k and k["name"] == app:
                options = k["options"] if "options" in k else {}
                break
    return options


# -----

# TODO
def __make_class_decorator(function_decorator):
    """
    Turn a function decorator, to a class decorator. It applies on individual
    method too
    :param function_decorator:
    :return:

    @make_class_decorator
    def function_decorator(fn):
        def gn(*args, **kwargs):
            print 'hi'
            fn(*args, **kwargs)
        gn.__name__ = fn.__name__
        return gn

    @function_decorator
    class Foo:
        def f(self):
            print 'hey'
    """
    def class_decorator(cls):
        if inspect.isclass(cls):
            for name, method in get_interesting_members(Shaft, cls):
                setattr(cls, name, function_decorator(method))
        else:
            return function_decorator(cls)
        return cls
    return class_decorator



import datetime
import uuid
import arrow
import exceptions
from flask_login import fresh_login_required
from shaft import (Shaft,
                   models,
                   page_meta,
                   init_app,
                   get_config,
                   session,
                   request,
                   redirect,
                   url_for,
                   flash_success,
                   flash_error,
                   abort,
                   recaptcha,
                   storage,
                   utils,
                   decorators as shaft_deco,
                   utils)
from . import (current_user,
               register,
               authenticated,
               logout_user,
               is_authenticated,
               not_authenticated,
               require_login_allowed,
               require_register_allowed,
               require_social_login_allowed,
               send_password_reset_email,
               send_email_verification_email,
               send_registration_welcome_email,
               session_set_require_password_change)

def main(**kwargs):

    options = kwargs.get("options", {})
    nav_title_partial = "AuthAccount/nav_title_partial.html"
    nav_kwargs = kwargs.get("nav_menu", {})
    verify_email = options.get("verify_email") or False

    @shaft_deco.menu_title(
        title=nav_kwargs.pop("title", "My Account") or "My Account",
        visible=is_authenticated,
        css_id=nav_kwargs.pop("css_id", "auth-account-menu"),
        css_class=nav_kwargs.pop("css_class", "auth-account-menu"),
        align_right=nav_kwargs.pop("align_right", True),
        title_partial=nav_kwargs.pop("title_partial", nav_title_partial),
        **nav_kwargs)
    class AuthAccount(Shaft):
        base_route = kwargs.get("route") or "/"
        decorators = Shaft.decorators + [authenticated] + kwargs.get("decorators")

        def _confirm_password(self):
            user_login = current_user.user_login("email")
            password = request.form.get("confirm-password")
            if not user_login.password_matched(password):
                raise exceptions.AuthError("Invalid password")
            return True

        @shaft_deco.menu_title("Logout", endpoint="AuthLogin:logout", order=100)
        def _(self): pass

        @shaft_deco.menu_title("Account Info", order=1)
        def account_info(self):
            page_meta(title="Account Info")

        @shaft_deco.menu_title("Edit Account Info", visible=False)
        @shaft_deco.accept_post_get
        def edit_account_info(self):
            page_meta(title="Edit Account Info")

            if request.method == "POST":
                try:
                    data = {
                        "name": request.form.get("name"),
                        "first_name": request.form.get("first_name"),
                        "last_name": request.form.get("last_name"),
                        "address": request.form.get("address"),
                        "address_city": request.form.get("address_city"),
                        "address_state": request.form.get("address_state"),
                        "address_country": request.form.get("address_country"),
                        "address_zip_code": request.form.get("address_zip_code"),
                        "telephone": request.form.get("telephone"),
                    }
                    dob = request.form.get("dob")
                    if dob:
                        data["date_of_birth"] = arrow.get(dob, models.AuthUser.DOB_FORMAT).datetime
                    current_user.update(**data)

                    flash_success("Account info updated successfully!")

                except exceptions.AuthError as ae:
                    flash_error(ae.message)
                except Exception as e:
                    logging.exception(e)
                    flash_error("Unable to edit account info")
                return redirect(self.account_info)

        @shaft_deco.menu_title("Change Login", order=2)
        @shaft_deco.accept_post_get
        def change_email(self):
            page_meta(title="Change Login Email")

            user_login = current_user.user_login("email")
            if not user_login:
                abort(400, "Require and Email account login to continue")

            if request.method == "POST":
                try:
                    self._confirm_password()

                    email = request.form.get("email")
                    if utils.is_email_valid(email):
                        user_login.change_email(email)
                        flash_success("Login Email changed successfully!")
                        return redirect(self.account_info)
                    else:
                        raise exceptions.AuthError("Invalid email")
                except exceptions.AuthError as ex:
                    flash_error("Error: %s" % ex)
                    return redirect(self)
                except Exception as e:
                    logging.exception(e)
                    flash_error("Unable to change email")
                    return redirect(self)

            return {
                "user_login": user_login
            }

        @shaft_deco.menu_title("Change Password", order=3)
        @shaft_deco.accept_post_get
        def change_password(self):
            page_meta(title="Change Password")

            user_login = current_user.user_login("email")
            if not user_login:
                abort(400, "Require and Email account login to continue")

            if request.method == "POST":
                try:
                    self._confirm_password()
                    password = request.form.get("password")
                    password2 = request.form.get("password2")
                    if password != password2:
                        raise exceptions.AuthError("Passwords don't match")
                    elif utils.is_password_valid(password):
                        user_login.change_password(password.strip())
                        session_set_require_password_change(False)
                        flash_success("Password updated successfully!")
                        return redirect(self.account_info)
                    else:
                        raise exceptions.AuthError("Invalid password")
                except exceptions.AuthError as ex:
                    flash_error("Error: %s" % ex)
                    return redirect(self.change_password)
                except Exception as e:
                    logging.exception(e)
                    flash_error("Unable to change password")
                    return redirect(self)

        @shaft_deco.menu_title("Change Profile Image", order=4)
        @shaft_deco.accept_post_get
        def change_profile_image(self):
            page_meta(title="Change Profile Image")

            if request.method == "POST":
                file = request.files.get("file")
                if file:
                    name = "%s" % uuid.uuid4().hex
                    extensions = ["jpg", "jpeg", "png", "gif"]
                    my_photo = storage.upload(file,
                                              name=name,
                                              prefix="profile-images/",
                                              allowed_extensions=extensions)
                    if my_photo:
                        url = my_photo.url
                        current_user.update(profile_image_url=url)
                        flash_success("Profile Picture updated successfully!")
                return redirect(self.account_info)

        @shaft_deco.menu_title("Setup Login", visible=False)
        @shaft_deco.accept_post_get
        def setup_login(self):
            user_login = current_user.user_login("email")
            if user_login:
                return redirect(self.account_info)

            if request.method == "POST":
                try:
                    email = request.form.get("email")
                    password = request.form.get("password")
                    password2 = request.form.get("password2")

                    if not password.strip() or password.strip() != password2.strip():
                        raise exceptions.AuthError("Passwords don't match")
                    else:
                        new_login = models.AuthUserLogin.new(login_type="email",
                                                             user_id=current_user.id,
                                                             email=email,
                                                             password=password.strip())
                        if verify_email:
                            send_registration_welcome_email(new_login.user)
                            flash_success("A welcome email containing a confirmation link has been sent to your email")
                            return redirect(self.account_info)
                except exceptions.AuthError as ex:
                    flash_error(ex.message)
                    return redirect(self.setup_login)
                except Exception as e:
                    logging.exception(e)
                    flash_error("Unable to setup login")
                    return redirect(self)

    return AuthAccount






"""
Shaft

Error

"""


class ShaftError(Exception):
    """
    This exception is not reserved, but it used for all Shaft exception.
    It helps catch Core problems.
    """
    pass


class AppError(ShaftError):
    """
    Use this exception in your application level.
    """
    pass


class ModelError(AppError):
    """
    Use this exception in your model level.
    """
    pass


class ExtensionError(ShaftError):
    """
    This Exception wraps the exception that was raised.
    Having multiple extension, we need a way to catch them all :)
    """
    def __init__(self, exc):
        """
        :param exc: Exception
        """
        self.exception = exc
        message = "%s : %s" % (exc.__class__.__name__, exc.message)
        super(self.__class__, self).__init__(message)











"""
Shaft

Views

"""
from shaft import (Shaft,
                   page_meta,
                   get_config,
                   flash_success,
                   flash_error,
                   abort,
                   request,
                   url_for,
                   redirect,
                   models,
                   utils,
                   decorators as deco
                   )

# ------------------------------------------------------------------------------


class Index(Shaft):

    @deco.menu_title("Home", order=1)
    def index(self):
        page_meta(title="Hello View!")
        return {}

    @deco.accept_get
    @deco.render_json
    def hello(self):
        return {"Hello": "World"}


